import collections
import dataclasses
import json
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any, Union, Tuple

import conllu
from overrides import overrides


@dataclass
class Token:
    id: Optional[Union[int, Tuple]] = None
    token: Optional[str] = None
    lemma: Optional[str] = None
    upostag: Optional[str] = None
    xpostag: Optional[str] = None
    feats: Optional[str] = None
    head: Optional[int] = None
    deprel: Optional[str] = None
    deps: Optional[str] = None
    misc: Optional[str] = None
    semrel: Optional[str] = None
    embeddings: Dict[str, List[float]] = field(default_factory=list, repr=False)



@dataclass
class Sentence:
    tokens: List[Token] = field(default_factory=list)
    sentence_embedding: List[float] = field(default_factory=list, repr=False)
    relation_distribution: List[float] = field(default_factory=list, repr=False)
    relation_label_distribution: List[float] = field(default_factory=list, repr=False)
    metadata: Dict[str, Any] = field(default_factory=collections.OrderedDict)

    def to_json(self):
        return json.dumps({
            "tokens": [dataclasses.asdict(t) for t in self.tokens],
            "sentence_embedding": self.sentence_embedding,
            "relation_distribution": self.relation_distribution,
            "relation_label_distribution": self.relation_label_distribution,
            "metadata": self.metadata,
        })

    def __len__(self):
        return len(self.tokens)


class _TokenList(conllu.TokenList):

    @overrides
    def __repr__(self):
        return 'TokenList<' + ', '.join(token['token'] for token in self) + '>'


def sentence2conllu(sentence: Sentence, keep_semrel: bool = False) -> conllu.TokenList:
    tokens = []
    for token in sentence.tokens:
        token_dict = collections.OrderedDict(dataclasses.asdict(token))
        # Remove semrel to have default conllu format.
        if not keep_semrel:
            del token_dict["semrel"]
        del token_dict["embeddings"]
        tokens.append(token_dict)
    # Range tokens must be tuple not list, this is conllu library requirement
    for t in tokens:
        if type(t["id"]) == list:
            t["id"] = tuple(t["id"])
        if t["deps"]:
            for dep in t["deps"]:
                if len(dep) > 1 and type(dep[1]) == list:
                    dep[1] = tuple(dep[1])
    return _TokenList(tokens=tokens,
                      metadata=sentence.metadata)


def tokens2conllu(tokens: List[str]) -> conllu.TokenList:
    return _TokenList(
        [collections.OrderedDict({"id": idx, "token": token}) for
         idx, token
         in enumerate(tokens, start=1)],
        metadata=collections.OrderedDict()
    )


def conllu2sentence(conllu_sentence: conllu.TokenList,
                    sentence_embedding=None, embeddings=None, relation_distribution=None,
                    relation_label_distribution=None) -> Sentence:
    if relation_distribution is None:
        relation_distribution = []
    if relation_label_distribution is None:
        relation_label_distribution = []
    if embeddings is None:
        embeddings = {}
    if sentence_embedding is None:
        sentence_embedding = []
    tokens = []
    for token in conllu_sentence.tokens:
        tokens.append(
            Token(
                **token, embeddings=embeddings[token["id"]]
            )
        )
    return Sentence(
        tokens=tokens,
        sentence_embedding=sentence_embedding,
        relation_distribution=relation_distribution,
        relation_label_distribution=relation_label_distribution,
        metadata=conllu_sentence.metadata
    )
